/*
Copyright The ORAS Authors.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package utils

var (
	Flags = struct {
		Layout           string
		FromLayout       string
		ToLayout         string
		FromLayoutPath   string
		ToLayoutPath     string
		DistributionSpec string
		ImageSpec        string
		IncludeReferrers string
		ExcludeReferrers string
	}{
		"--oci-layout",
		"--from-oci-layout",
		"--to-oci-layout",
		"--from-oci-layout-path",
		"--to-oci-layout-path",
		"--distribution-spec",
		"--image-spec",
		"--include-referrers",
		"--exclude-referrers",
	}
	RegistryErrorPrefix = "Error response from registry: "
	InvalidTag          = "i-dont-think-this-tag-exists"
)
